from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.db.models import Count
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_constants.constants import NO, PARTIAL, YES
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...forms import StockTransferForm
from ...models import ConfirmationAtSiteItem, StockTransfer, StockTransferItem
from ..actions import print_transfer_stock_manifest_action, transfer_stock_action
from ..model_admin_mixin import ModelAdminMixin


class ConfirmedAtSiteListFilter(SimpleListFilter):
    title = "Confirmed at site"
    parameter_name = "confirmed_at_site"

    def lookups(self, request, model_admin):
        return (YES, YES), (PARTIAL, "Partial"), (NO, NO)

    def queryset(self, request, queryset):
        qs = None
        if self.value():
            if self.value() == YES:
                qs = (
                    queryset.filter(
                        confirmationatsite__isnull=False,
                        stocktransferitem__stock__confirmationatsiteitem__isnull=False,
                    )
                    .exclude(
                        stocktransferitem__stock__confirmationatsiteitem__isnull=True,
                    )
                    .annotate(Count("transfer_identifier"))
                )
            elif self.value() == PARTIAL:
                qs = queryset.filter(
                    confirmationatsite__isnull=False,
                    stocktransferitem__stock__confirmationatsiteitem__isnull=True,
                ).annotate(Count("transfer_identifier"))
            elif self.value() == NO:
                qs = queryset.filter(
                    confirmationatsite__isnull=True,
                    stocktransferitem__stock__confirmationatsiteitem__isnull=True,
                ).annotate(Count("transfer_identifier"))

        return qs


@admin.register(StockTransfer, site=edc_pharmacy_admin)
class StockTransferAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock Transfer"
    change_form_title = "Pharmacy: Stock Transfers"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20
    ordering = ["-transfer_identifier"]

    autocomplete_fields = ["from_location", "to_location"]
    actions = [transfer_stock_action, print_transfer_stock_manifest_action]

    form = StockTransferForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "transfer_identifier",
                    "transfer_datetime",
                    "from_location",
                    "to_location",
                    "item_count",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "transfer_date",
        "from_location",
        "to_location",
        "stock_transfer_item_changelist",
        "stock_transfer_item_confirmed_changelist",
        "stock_transfer_item_unconfirmed_changelist",
        "stock_changelist",
    )

    list_filter = (
        ("transfer_datetime", DateRangeFilterBuilder()),
        "from_location",
        "to_location",
        ConfirmedAtSiteListFilter,
    )

    search_fields = (
        "id",
        "transfer_identifier",
        "stocktransferitem__stock__allocation__registered_subject__subject_identifier",
    )

    @admin.display(description="TRANSFER #", ordering="transfer_identifier")
    def identifier(self, obj):
        return obj.transfer_identifier

    @admin.display(description="Transfer date", ordering="transfer_datetime")
    def transfer_date(self, obj):
        return to_local(obj.transfer_datetime).date()

    @admin.display(description="Transfered")
    def stock_transfer_item_changelist(self, obj):
        count = StockTransferItem.objects.filter(stock_transfer=obj).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label=count, title="Go to stock transfer items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Confirmed at site")
    def stock_transfer_item_confirmed_changelist(self, obj):
        num_confirmed_at_site = ConfirmationAtSiteItem.objects.filter(
            stock__stocktransferitem__stock_transfer=obj
        ).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}&confirmed_at_site={YES}"
        context = dict(
            url=url,
            label=num_confirmed_at_site,
            title="Items confirmed at site",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Unconfirmed")
    def stock_transfer_item_unconfirmed_changelist(self, obj):
        num_transferred = StockTransferItem.objects.filter(
            stock_transfer=obj, stock__confirmationatsiteitem__isnull=False
        ).count()
        num_confirmed_at_site = ConfirmationAtSiteItem.objects.filter(
            stock__stocktransferitem__stock_transfer=obj
        ).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}&confirmed_at_site={NO}"
        context = dict(
            url=url,
            label=num_transferred - num_confirmed_at_site,
            title="Items not confirmed at site",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock", ordering="stock__code")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



# Generated by Django 3.0.4 on 2020-03-27 11:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0015_auto_20200227_2327"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
    ]



from __future__ import annotations

from django.utils.text import slugify
from django_crypto_fields.utils import get_encrypted_field_names

SLUG_SEP = "|"


def generate_slug(obj, fields) -> str | None:
    """Returns a slug of the char values of the fields
    listed on the models get_search_slug_fields() method.

    Excludes any encrypted fields if listed.
    """
    slug = None
    if obj and fields:
        values = []
        for field in (f for f in fields if f not in get_encrypted_field_names(obj)):
            v = obj
            for f in field.split("."):
                v = getattr(v, f)
            if isinstance(v, str):
                values.append(v[:50])  # truncate value
        slugs = [slugify(v) for v in list(set(values)) if v]
        slug = SLUG_SEP.join(slugs)[:250]
    return slug



from __future__ import annotations

import contextlib
import uuid
from uuid import uuid4

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView

from edc_constants.constants import CONFIRMED
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..constants import ALREADY_CONFIRMED, INVALID
from ..models import ConfirmationAtSite, Location, Stock, StockTransfer
from ..utils import confirm_stock_at_site


@method_decorator(login_required, name="dispatch")
class ConfirmationAtSiteView(
    EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView
):
    template_name: str = "edc_pharmacy/stock/confirmation_at_site.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"

    def __init__(self, **kwargs):
        self.session_uuid: str | None = None
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        extra_opts = {}
        stock_transfer = self.get_stock_transfer(self.kwargs.get("stock_transfer_identifier"))
        if not self.kwargs.get("session_uuid"):
            self.session_uuid = str(uuid4())
            session_obj = None
        else:
            self.session_uuid = str(self.kwargs.get("session_uuid"))
            session_obj = self.request.session[self.session_uuid]
        if stock_transfer:
            unconfirmed_count = self.get_unconfirmed_count(stock_transfer)
            extra_opts = dict(
                unconfirmed_count=unconfirmed_count,
                item_count=list(
                    range(1, self.get_adjusted_unconfirmed_count(stock_transfer) + 1)
                ),
                last_codes=[],
            )
        if session_obj:
            last_codes = [(x, "confirmed") for x in session_obj.get("confirmed") or []]
            last_codes.extend(
                [(x, "already confirmed") for x in session_obj.get("already_confirmed") or []]
            )
            last_codes.extend([(x, "invalid") for x in session_obj.get("invalid") or []])
            unconfirmed_count = self.get_unconfirmed_count(stock_transfer)
            extra_opts.update(
                item_count=list(
                    range(1, self.get_adjusted_unconfirmed_count(stock_transfer) + 1)
                ),
                unconfirmed_count=unconfirmed_count,
                last_codes=last_codes,
            )
        kwargs.update(
            locations=Location.objects.filter(
                site__in=self.request.user.userprofile.sites.all()
            ),
            location=self.location,
            location_id=self.location_id,
            stock_transfer=stock_transfer,
            stock_transfers=self.stock_transfers,
            session_uuid=str(self.session_uuid),
            CONFIRMED=CONFIRMED,
            ALREADY_CONFIRMED=ALREADY_CONFIRMED,
            INVALID=INVALID,
            **extra_opts,
        )
        return super().get_context_data(**kwargs)

    @property
    def stock_transfers(self):
        qs = StockTransfer.objects.filter(
            to_location__site=self.site,
            stocktransferitem__stock__confirmationatsiteitem__isnull=True,
        )
        return qs.annotate(count=Count("transfer_identifier")).order_by("-transfer_datetime")

    def get_adjusted_unconfirmed_count(self, stock_transfer):
        unconfirmed_count = self.get_unconfirmed_count(stock_transfer)
        return min(unconfirmed_count, 12)

    def get_stock_codes(self, stock_transfer):
        return [
            code
            for code in stock_transfer.stocktransferitem_set.values_list(
                "stock__code", flat=True
            ).all()
        ]

    def get_unconfirmed_count(self, stock_transfer) -> int:
        return (
            Stock.objects.values("code")
            .filter(
                code__in=self.get_stock_codes(stock_transfer),
                location_id=self.location_id,
                confirmationatsiteitem__isnull=True,
            )
            .count()
        )

    @property
    def site(self) -> Site | None:
        obj = None
        if self.kwargs.get("site_id"):
            with contextlib.suppress(ObjectDoesNotExist):
                obj = Site.objects.get(id=self.kwargs.get("site_id"))
        return obj

    @property
    def location_id(self) -> uuid.UUID | None:
        location_id = self.kwargs.get("location_id") or self.request.POST.get("location_id")
        if not location_id and self.site:
            try:
                location = Location.objects.get(site=self.site)
            except ObjectDoesNotExist:
                pass
            else:
                location_id = location.id
        return location_id

    @property
    def location(self) -> Location:
        try:
            location = Location.objects.get(pk=self.location_id)
        except ObjectDoesNotExist:
            location = None
        return location

    @property
    def stock_codes(self) -> list[str]:
        session_uuid = self.kwargs.get("session_uuid")
        if session_uuid:
            return self.request.session[str(session_uuid)].get("stock_codes")
        return []

    @property
    def confirmation_at_site(self):
        confirmation_at_site_id = self.kwargs.get("confirmation_at_site")
        try:
            confirmation_at_site = ConfirmationAtSite.objects.get(id=confirmation_at_site_id)
        except ObjectDoesNotExist:
            confirmation_at_site = None
            messages.add_message(
                self.request, messages.ERROR, "Invalid stock transfer confirmation."
            )
        return confirmation_at_site

    @property
    def confirmation_at_site_changelist_url(self) -> str:
        if self.confirmation_at_site:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_confirmationatsite_changelist")
            return f"{url}?q={self.confirmation_at_site.transfer_confirmation_identifier}"
        return "/"

    def get_stock_transfer(
        self,
        stock_transfer_identifier: str,
        suppress_msg: bool | None = None,
    ) -> StockTransfer | None:
        stock_transfer = None
        try:
            stock_transfer = StockTransfer.objects.get(
                transfer_identifier=stock_transfer_identifier or None,
                to_location_id=self.location_id or None,
            )
        except ObjectDoesNotExist:
            if stock_transfer_identifier and not suppress_msg:
                location = Location.objects.get(pk=self.location_id or None)
                messages.add_message(
                    self.request,
                    messages.ERROR,
                    (
                        "Invalid Reference. Please check the manifest "
                        "reference and delivery site. "
                        f"Got {stock_transfer_identifier} at {location}."
                    ),
                )
        return stock_transfer

    def post(self, request, *args, **kwargs) -> HttpResponseRedirect:  # noqa: ARG002
        # cancel
        if request.POST.get("cancel") and request.POST.get("cancel") == "cancel":
            url = reverse("edc_pharmacy:home_url")
            return HttpResponseRedirect(url)

        stock_transfer_identifier = request.POST.get("stock_transfer_identifier")
        stock_transfer = self.get_stock_transfer(stock_transfer_identifier, suppress_msg=True)
        location_id = request.POST.get("location_id")
        if not stock_transfer or not location_id:
            # nothing selected
            url = reverse("edc_pharmacy:confirmation_at_site_url", kwargs={})
            return HttpResponseRedirect(url)

        session_uuid = request.POST.get("session_uuid")
        stock_codes = (
            request.POST.getlist("stock_codes") if request.POST.get("stock_codes") else []
        )

        # you have unconfirmed items, so go to the scan page
        if not stock_codes and stock_transfer.uncomfirmed_items > 0:
            url = reverse(
                "edc_pharmacy:confirmation_at_site_url",
                kwargs={
                    "stock_transfer_identifier": stock_transfer_identifier,
                    "location_id": location_id,
                    "items_to_scan": stock_transfer.uncomfirmed_items,
                },
            )
            return HttpResponseRedirect(url)

        if stock_codes:
            # you have scanned codes, process them
            confirmed, already_confirmed, invalid = confirm_stock_at_site(
                stock_transfer, stock_codes, location_id, request=request
            )
            # message for confirmed
            if confirmed:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    f"Successfully confirmed {len(confirmed)} stock items. ",
                )
            # message for already confirmed
            if already_confirmed:
                messages.add_message(
                    request,
                    messages.WARNING,
                    (
                        f"Skipped {len(already_confirmed)} items. Stock items are "
                        "already confirmed."
                    ),
                )
            # message for invalid codes
            if invalid:
                messages.add_message(
                    request,
                    messages.ERROR,
                    f"Invalid codes submitted! Got {', '.join(invalid)} .",
                )

            self.request.session[session_uuid] = dict(
                confirmed=confirmed,
                already_confirmed=already_confirmed,
                invalid=invalid,
                stock_transfer_pk=str(stock_transfer.pk),
            )
            # return to page with any unconfirmed codes for this stock_transfer document
            # might be 0 items
            url = reverse(
                "edc_pharmacy:confirmation_at_site_url",
                kwargs={
                    "session_uuid": str(request.POST.get("session_uuid")),
                    "stock_transfer_identifier": stock_transfer_identifier,
                    "location_id": location_id,
                    "items_to_scan": stock_transfer.uncomfirmed_items,
                },
            )
            return HttpResponseRedirect(url)
        # can you get here??
        return HttpResponseRedirect(self.confirmation_at_site_changelist_url)



from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.db.models import DateField, DateTimeField


class InlineModelFormMixinError(Exception):
    pass


class InlineModelFormMixin:
    def get_inline_field_values(self, field=None, inline_set=None):
        """Returns a list values from the inline.

        The value of fk fields is the id."""
        field_values = []
        total_forms = int(self.data.get(f"{inline_set}-TOTAL_FORMS"))
        for i in range(0, total_forms):
            if self.data.get(f"{inline_set}-{i}-DELETE") == "on":
                pass
            else:
                inline_field = f"{inline_set}-{i}-{field}"
                try:
                    field_values.append(self.data[inline_field].name)
                except KeyError:
                    break
                except AttributeError:
                    field_values.append(self.data[inline_field])
        return field_values

    def unique_inline_values_or_raise(self, field=None, inline_model=None, field_label=None):
        self.field_exists_or_raise(field, inline_model)
        inline_set = f"{inline_model.split('.')[1]}_set"
        items = self.get_inline_field_values(field=field, inline_set=inline_set)
        if len(items) != len(list(set(items))):
            field_label = field_label or field
            raise forms.ValidationError(
                f"{field_label}: The list of values below must be unique"
            )

    def dates_not_after_report_datetime(self, field=None, inline_model=None, field_label=None):
        self.field_cls_is_date_or_raise(field=field, inline_model=inline_model)
        inline_set = f"{inline_model.split('.')[1]}_set"
        dates_as_str = self.get_inline_field_values(field=field, inline_set=inline_set)
        for dte_as_str in dates_as_str:
            if dte_as_str:
                try:
                    dte = datetime.fromisoformat(dte_as_str)
                except ValueError as e:
                    raise forms.ValidationError(
                        f"{field_label}: Invalid date or date format. Got {dte_as_str}"
                    ) from e
                else:
                    if dte.astimezone(ZoneInfo(settings.TIME_ZONE)) > self.cleaned_data.get(
                        "report_datetime"
                    ):
                        raise forms.ValidationError(
                            f"{field_label}: Date cannot be after report date/time. "
                            f"Got `{dte_as_str}`."
                        )

    def field_exists_or_raise(self, field, inline_model):
        model_cls = django_apps.get_model(inline_model)
        if field not in [f.name for f in model_cls._meta.get_fields()]:
            raise InlineModelFormMixinError(
                f"Field does not exist on model class. See {self.__class__.__name__}. "
                f"Got {inline_model}.{field}."
            )
        return [f for f in model_cls._meta.get_fields() if f.name == field][0]

    def field_cls_is_date_or_raise(self, field: str, inline_model) -> Any:
        """Raises if field class on model is not a date.

        Works for DateField and DatetimeField"""
        field_obj = self.field_exists_or_raise(field, inline_model)
        if field_obj.__class__ not in [DateTimeField, DateField]:
            raise InlineModelFormMixinError(
                f"Field is not a date field class. See {self.__class__.__name__}. "
                f"Got {inline_model}.{field}."
            )
        return field_obj.__class__



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist

from edc_consent.form_validators import ConsentDefinitionFormValidatorMixin
from edc_form_validators import ReportDatetimeFormValidatorMixin
from edc_utils import age, to_utc
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit_schedule import VisitSchedule
from edc_visit_tracking.modelform_mixins import get_related_visit

if TYPE_CHECKING:
    from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
    from edc_model.models import BaseUuidModel
    from edc_sites.model_mixins import SiteModelMixin
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    class RelatedVisitModel(SiteModelMixin, CreatesMetadataModelMixin, Base, BaseUuidModel):
        pass


__all__ = ["BaseFormValidatorMixin", "CrfFormValidatorMixin"]


class BaseFormValidatorMixin(
    ReportDatetimeFormValidatorMixin, ConsentDefinitionFormValidatorMixin
):
    """A base mixin of common properties needed for PRN/CRF validation
    to be declared with FormValidators.
    """

    @property
    def subject_identifier(self) -> str | None:
        if "subject_identifier" in self.cleaned_data:
            subject_identifier = self.cleaned_data.get("subject_identifier")
        else:
            subject_identifier = getattr(self.instance, "subject_identifier", None)
        return subject_identifier

    @property
    def age_in_years(self) -> int | None:
        if self.report_datetime and self.subject_consent.dob:
            return age(self.subject_consent.dob, to_utc(self.report_datetime)).years
        return None


class CrfFormValidatorMixin(BaseFormValidatorMixin):
    """Assumes model is a CRF and has a key to related/subject_visit.

    Declare with FormValidator (not modelform).
    """

    @property
    def subject_identifier(self) -> str:
        """Always returns the subject_identifier from related_visit"""
        return self.related_visit.subject_identifier

    @property
    def report_datetime(self) -> datetime:
        """Returns report_datetime or raises.

        Report datetime is always a required field on a CRF model,
        Django will raise a field ValidationError before getting
        here if report_datetime is None.
        """
        report_datetime = None
        if self.report_datetime_field_attr in self.cleaned_data:
            report_datetime = self.cleaned_data.get(self.report_datetime_field_attr)
        elif self.instance:
            report_datetime = self.instance.report_datetime
        return report_datetime

    @property
    def related_visit_model_attr(self) -> str:
        return self.model.related_visit_model_attr()

    @property
    def related_visit(self) -> RelatedVisitModel:
        """Returns a subject visit model instance or None"""
        return get_related_visit(self, related_visit_model_attr=self.related_visit_model_attr)

    @property
    def related_visit_model_cls(self) -> type[RelatedVisitModel]:
        """Returns a subject visit model instance or None"""
        return get_related_visit(
            self, related_visit_model_attr=self.related_visit_model_attr
        ).__class__

    @property
    def visit_schedule(self) -> VisitSchedule:
        return self.related_visit.visit_schedule

    @property
    def schedule(self) -> Schedule:
        return self.related_visit.schedule

    @property
    def offschedule_datetime(self) -> datetime | None:
        try:
            obj = self.related_visit.schedule.offschedule_model_cls.objects.get(
                subject_identifier=self.subject_identifier
            )
        except ObjectDoesNotExist:
            offschedule_datetime = None
        else:
            offschedule_datetime = obj.offschedule_datetime
        return offschedule_datetime



import sys

from .site_randomizers import site_randomizers

RANDO_UNBLINDED = "RANDO_UNBLINDED"
RANDO_BLINDED = "RANDO_BLINDED"


def get_rando_permissions_tuples() -> list:
    """Returns a list of tuples, [(codename, description), ...]"""
    rando_tuples = []
    for randomizer_cls in site_randomizers._registry.values():
        app_label, model = randomizer_cls.model_cls()._meta.label_lower.split(".")
        verbose_name = randomizer_cls.model_cls()._meta.verbose_name
        rando_tuples.append((f"{app_label}.view_{model}", f"Can view {verbose_name}"))
        if not randomizer_cls.trial_is_blinded:
            rando_tuples.append(
                (
                    f"{app_label}.display_{model}",
                    f"Can display {verbose_name} assignment",
                )
            )
    return rando_tuples


def get_rando_permissions_codenames() -> list:
    """Returns a list of codenames"""
    return [c[0] for c in get_rando_permissions_tuples()]


# codenames
export_rando = [
    "edc_randomization.export_randomizationlist",
]


# post_update_func
def update_rando_group_permissions(auth_updater, app_label: str | None):
    """Update group permissions for each registered randomizer class."""
    for randomizer_cls in site_randomizers._registry.values():
        if auth_updater.verbose:
            sys.stdout.write(
                "     - creating permissions for registered randomizer_cls "
                f"`{randomizer_cls.name}` model "
                f"`{randomizer_cls.model_cls()._meta.label_lower}`\n"
            )
        rando_tuples = [
            (k, v)
            for k, v in get_rando_permissions_tuples()
            if k.startswith(randomizer_cls.model_cls()._meta.label_lower.split(".")[0])
        ]
        auth_updater.group_updater.create_permissions_from_tuples(
            randomizer_cls.model_cls()._meta.label_lower,
            rando_tuples,
        )


def make_randomizationlist_view_only(auth_updater, app_label: str | None):
    for randomizer_cls in site_randomizers._registry.values():
        randomizer_cls.apps = auth_updater.apps
        app_label, model = randomizer_cls.model_cls()._meta.label_lower.split(".")
        permissions = auth_updater.group_updater.permission_model_cls.objects.filter(
            content_type__app_label=app_label, content_type__model=model
        ).exclude(codename=f"view_{model}")
        codenames = [f"{app_label}.{o.codename}" for o in permissions]
        codenames.extend(
            [
                f"{app_label}.add_{model}",
                f"{app_label}.change_{model}",
                f"{app_label}.delete_{model}",
            ]
        )
        codenames = list(set(codenames))
        for group in auth_updater.group_updater.group_model_cls.objects.all():
            auth_updater.group_updater.remove_permissions_by_codenames(
                group=group,
                codenames=codenames,
                allow_multiple_objects=True,
            )



import pandas as pd

from ..table import Table


class GenderTable(Table):
    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(
            colname="gender",
            main_df=main_df,
            title="Gender",
        )



from django.conf import settings

from edc_navbar import Navbar, NavbarItem, site_navbars

no_url_namespace = True if settings.APP_NAME == "edc_adverse_event" else False

ae_navbar_item = NavbarItem(
    name="ae_home",
    title="Adverse Events",
    label="AE",
    codename="edc_adverse_event.nav_ae_section",
    url_name="edc_adverse_event:ae_home_url",
    no_url_namespace=no_url_namespace,
)

tmg_navbar_item = NavbarItem(
    name="tmg_home",
    label="TMG",
    codename="edc_adverse_event.nav_tmg_section",
    url_name="edc_adverse_event:tmg_home_url",
    no_url_namespace=no_url_namespace,
)

ae_navbar = Navbar(name="edc_adverse_event")
ae_navbar.register(ae_navbar_item)
ae_navbar.register(tmg_navbar_item)

site_navbars.register(ae_navbar)



from django.contrib import admin, messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext


@admin.display(description="Repack stock")
def go_to_add_repack_request_action(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    elif not getattr(queryset.first(), "confirmation", None):
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Unable to repack. Stock item has not been confirmed."),
        )
    else:
        obj = queryset.first()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_repackrequest_add")
        url = (
            f"{url}?next=edc_pharmacy_admin:edc_pharmacy_repackrequest_changelist,q"
            f"&q={obj.code}&from_stock={obj.id}"
        )
        return HttpResponseRedirect(url)
    return None



# Generated by Django 5.1.2 on 2024-11-03 17:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pylabels", "0002_alter_label_options_label_created_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="label",
            name="sample_data",
            field=models.TextField(
                blank=True,
                help_text="for example, if the label_data cls is a model",
                null=True,
            ),
        ),
    ]



from typing import Any

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.views import LoginView as BaseLoginView
from django.contrib.auth.views import LogoutView as BaseLogoutView
from django_revision.revision import site_revision


class LoginView(BaseLoginView):
    template_name = "edc_auth/login.html"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        """Tests cookies."""
        kwargs.update(**self.extra_context)
        self.request.session.set_test_cookie()
        if not self.request.session.test_cookie_worked():
            messages.add_message(self.request, messages.ERROR, "Please enable cookies.")
        self.request.session.delete_test_cookie()
        return super().get_context_data(**kwargs)

    @property
    def extra_context(self):
        try:
            live_system = settings.LIVE_SYSTEM
        except AttributeError:
            live_system = None
        try:
            allow_password_reset = settings.ALLOW_PASSWORD_RESET
        except AttributeError:
            allow_password_reset = None
        return {
            "edc_base_template": "edc_dashboard/base.html",
            "DEBUG": settings.DEBUG,
            "ALLOW_PASSWORD_RESET": allow_password_reset,
            "revision": site_revision.tag,
            "live_system": live_system,
            "INDEX_PAGE": getattr(settings, "INDEX_PAGE", None),
            "INDEX_PAGE_LABEL": getattr(settings, "INDEX_PAGE_LABEL", None),
        }


class LogoutView(BaseLogoutView):
    template_name = "edc_auth/login.html"
    next_page = settings.LOGOUT_REDIRECT_URL or "accounts/login"

    @property
    def extra_context(self) -> dict:
        try:
            live_system = settings.LIVE_SYSTEM
        except AttributeError:
            live_system = None
        return {
            "DEBUG": settings.DEBUG,
            "edc_base_template": "edc_dashboard/base.html",
            "revision": site_revision.tag,
            "live_system": live_system,
            "INDEX_PAGE": getattr(settings, "INDEX_PAGE", None),
            "INDEX_PAGE_LABEL": getattr(settings, "INDEX_PAGE_LABEL", None),
        }



from __future__ import annotations

from ..database import Database
from ..df_handlers import DfHandler
from ..utils import get_export_folder, get_table_names
from .csv_exporter import CsvExporter


class TablesExporterError(Exception):
    pass


class InvalidTableName(Exception):
    pass


class TablesExporter:
    """Export to file all tables for an app_label.

    Does not decrypt values stored in django_crypto_fields
    encrypted field classes.

    Usage:
        exporter = TablesExporter(app_label='td')
        exporter.to_csv()
        exporter = TablesExporter(app_label='edc')
        exporter.to_csv()

    """

    app_label = None
    csv_exporter_cls = CsvExporter
    db_cls = Database
    delimiter = "|"
    df_handler_cls = DfHandler
    exclude_history_tables = False
    excluded_app_labels = ["django_collect_offline"]

    def __init__(
        self,
        app_label: str | None = None,
        with_columns: list[str] | None = None,
        without_columns: list[str] | None = None,
        exclude_history_tables: bool | None = None,
        exclude_table_hints: bool | None = None,
        export_folder: str | None = None,
        **kwargs,
    ):
        self.app_label = app_label or self.app_label
        if not self.app_label:
            raise TablesExporterError(f"Missing app_label. Got None. See {self!r}")
        self.export_folder = export_folder or get_export_folder()
        self.with_columns = with_columns or []
        self.without_columns = without_columns or []
        exclude_table_hints = exclude_table_hints or []
        self.exclude_history_tables = (
            self.exclude_history_tables
            if exclude_history_tables is None
            else exclude_history_tables
        )
        self.exported_paths = {}
        self.db = self.db_cls()
        self.table_names = self.get_table_names()
        if not self.table_names:
            raise TablesExporterError(
                f"No tables found for app_label. Got {self.app_label}. See {self!r}"
            )
        for hint in exclude_table_hints:
            for table_name in self.get_table_names():
                if hint in table_name:
                    self.table_names.pop(self.table_names.index(table_name))
        if self.exclude_history_tables:
            self.table_names = [
                tbl
                for tbl in self.table_names
                if "historical" not in tbl and not tbl.endswith("_audit")
            ]

    def __repr__(self):
        return f"{self.__class__.__name__}(app_label='{self.app_label}')"

    def to_csv(
        self,
        table_names: list[str] | None = None,
        export_folder: str | None = None,
        **kwargs,
    ):
        """Exports all tables to CSV as raw tables.

        See also `ModelToDataframe` for less generic export options.
        """
        self.exported_paths = {}
        export_folder = export_folder or self.export_folder
        if table_names:
            for table_name in table_names:
                if table_name not in self.table_names:
                    raise InvalidTableName(f"{table_name} is not a valid for {self}")
            self.table_names = table_names
        for table_name in self.table_names:
            df = self.to_df(table_name=table_name, **kwargs)
            exporter = self.csv_exporter_cls(
                table_name=table_name, export_folder=export_folder, **kwargs
            )
            exported = exporter.to_csv(dataframe=df, export_folder=export_folder)
            if exported.path:
                self.exported_paths.update({table_name: exported.path})

    def get_table_names(self) -> list[str]:
        """Returns a list of table names for this app_label."""
        return get_table_names(
            app_label=self.app_label,
            with_columns=self.with_columns,
            without_columns=self.without_columns,
            db_cls=self.db_cls,
        )

    def to_df(self, table_name: str = None, **kwargs):
        """Returns a dataframe after passing the raw df
        through the df_handler class.
        """
        df = self.get_raw_df(table_name)
        if self.df_handler_cls:
            df_handler = self.df_handler_cls(
                dataframe=df, db=self.db, table_name=table_name, **kwargs
            )
            df = df_handler.dataframe
        return df

    def get_raw_df(self, table_name: str = None):
        """Returns a df for the given table_name
        from an SQL statement, that is; raw).
        """
        return self.db.select_table(table_name=table_name)



from .aliquot_creator import AliquotCreator, AliquotCreatorError
from .aliquot_type import (
    AliquotType,
    AliquotTypeAlphaCodeError,
    AliquotTypeNumericCodeError,
)
from .lab_profile import (
    LabProfile,
    LabProfileRequisitionModelError,
    PanelAlreadyRegistered,
)
from .manifest import Manifest
from .primary_aliquot import PrimaryAliquot
from .processing_profile import (
    Process,
    ProcessingProfile,
    ProcessingProfileAlreadyAdded,
    ProcessingProfileInvalidDerivative,
)
from .requisition_panel import (
    InvalidProcessingProfile,
    RequisitionPanel,
    RequisitionPanelError,
    RequisitionPanelLookupError,
)
from .requisition_panel_group import RequisitionPanelGroup
from .specimen import Specimen, SpecimenNotDrawnError
from .specimen_processor import SpecimenProcessor, SpecimenProcessorError



from __future__ import annotations

from abc import ABC
from uuid import uuid4

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.utils import timezone
from django_revision.revision import Revision
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle, StyleSheet1, getSampleStyleSheet
from reportlab.lib.units import cm
from reportlab.platypus import SimpleDocTemplate

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils.date import to_local

from .numbered_canvas import NumberedCanvas


class ReportError(Exception):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code


class Report(ABC):
    document_template = SimpleDocTemplate
    watermark_word: str | None = getattr(settings, "EDC_PDF_REPORTS_WATERMARK_WORD", None)
    watermark_font: tuple[str, int] | None = getattr(
        settings, "EDC_PDF_REPORTS_WATERMARK_FONT", ("Helvetica", 100)
    )
    default_numbered_canvas = NumberedCanvas

    default_page = dict(  # noqa: RUF012
        rightMargin=0.5 * cm,
        leftMargin=0.5 * cm,
        topMargin=1.5 * cm,
        bottomMargin=1.5 * cm,
        pagesize=A4,
    )

    def __init__(
        self,
        page: dict | None = None,
        header_line: str | None = None,
        filename: str | None = None,
        request: WSGIRequest | None = None,
        numbered_canvas: type[NumberedCanvas] | None = None,
        footer_row_height: int | None = None,
    ):
        self._styles = None
        self.request = request
        self.page = page or self.default_page
        self.filename = filename or f"{uuid4()}.pdf"
        self.footer_row_height = footer_row_height or 25
        self.numbered_canvas = numbered_canvas or self.default_numbered_canvas

        if not header_line:
            header_line = ResearchProtocolConfig().institution
        self.header_line = header_line

    def build(self, response):
        document_template = self.document_template(response, **self.page)
        story = self.get_report_story(document_template=document_template)  # flowables
        document_template.build(
            story,
            onFirstPage=self.on_first_page,
            onLaterPages=self.on_later_pages,
            canvasmaker=self.numbered_canvas,
        )

    @property
    def numbered_canvas(self):
        return self._numbered_canvas

    @numbered_canvas.setter
    def numbered_canvas(self, value: type[NumberedCanvas]):
        self._numbered_canvas = value
        if self.watermark_word:
            self._numbered_canvas.watermark_word = self.watermark_word
            if self.watermark_font:
                self._numbered_canvas.watermark_font = self.watermark_font

    @property
    def report_filename(self) -> str:
        return self.filename

    def get_report_story(self, **kwargs):
        """Entry point, returns a list of flowables to be passed to build.

        For example, where ``pdf_report`` is an instance of this class:

            buffer = BytesIO()
            doctemplate = pdf_report.document_template(buffer, **pdf_report.page)
            story = pdf_report.get_report_story()
            doctemplate.build(
                story,
                onFirstPage=pdf_report.on_first_page,
                onLaterPages=pdf_report.on_later_pages,
                canvasmaker=NumberedCanvas)
            buffer.seek(0)

        """
        return []

    def on_first_page(self, canvas, doc):
        """Callback for `onFirstPage`"""
        self.draw_footer(canvas, doc)

    def on_later_pages(self, canvas, doc):
        """Callback for onLaterPages"""
        self.draw_header(canvas, doc)
        self.draw_footer(canvas, doc)

    def draw_header(self, canvas, doc):
        pass

    def draw_footer(self, canvas, doc):  # noqa: ARG002
        styles = getSampleStyleSheet()
        styles.add(ParagraphStyle(name="header", fontSize=6, alignment=TA_CENTER))
        width, _ = A4
        canvas.setFontSize(6)
        timestamp = to_local(timezone.now()).strftime("%Y-%m-%d %H:%M")
        canvas.drawRightString(
            width - len(timestamp) - 20,
            self.footer_row_height,
            f"printed on {timestamp}",
        )
        canvas.drawString(35, self.footer_row_height, f"clinicedc {Revision().tag}")

    @property
    def styles(self):
        if not self._styles:
            styles = getSampleStyleSheet()
            styles.add(ParagraphStyle(name="titleR", fontSize=8, alignment=TA_RIGHT))
            styles.add(ParagraphStyle(name="header", fontSize=6, alignment=TA_CENTER))
            styles.add(ParagraphStyle(name="footer", fontSize=6, alignment=TA_RIGHT))
            styles.add(ParagraphStyle(name="center", alignment=TA_CENTER))
            styles.add(ParagraphStyle(name="right", alignment=TA_RIGHT))
            styles.add(ParagraphStyle(name="left", alignment=TA_LEFT))
            styles.add(
                ParagraphStyle(name="line_data", alignment=TA_LEFT, fontSize=8, leading=10)
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_small", alignment=TA_LEFT, fontSize=7, leading=9
                )
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_small_center",
                    alignment=TA_CENTER,
                    fontSize=7,
                    leading=8,
                )
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_medium", alignment=TA_LEFT, fontSize=10, leading=12
                )
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_mediumB",
                    alignment=TA_LEFT,
                    fontSize=10,
                    leading=11,
                    fontName="Helvetica-Bold",
                )
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_large", alignment=TA_LEFT, fontSize=11, leading=14
                )
            )
            styles.add(
                ParagraphStyle(
                    name="line_data_largest", alignment=TA_LEFT, fontSize=14, leading=18
                )
            )
            styles.add(
                ParagraphStyle(name="line_label", fontSize=7, leading=6, alignment=TA_LEFT)
            )
            styles.add(
                ParagraphStyle(name="line_label_center", fontSize=7, alignment=TA_CENTER)
            )
            styles.add(
                ParagraphStyle(name="row_header", fontSize=8, leading=8, alignment=TA_CENTER)
            )
            styles.add(
                ParagraphStyle(name="row_data", fontSize=7, leading=7, alignment=TA_CENTER)
            )
            self._styles = self.add_to_styles(styles)
        return self._styles

    def add_to_styles(self, styles: StyleSheet1) -> StyleSheet1:
        return styles



from django.contrib.auth.models import User as BaseUser


class DataManagerUser(BaseUser):
    def __str__(self):
        return f"{self.first_name} {self.last_name} ({self.username})"

    class Meta:
        proxy = True
        default_permissions = ("view", "export")
        verbose_name = "Data Manager User"
        verbose_name_plural = "Data Manager Users"


class QueryUser(BaseUser):

    # see edc_model_to_dataframe
    m2m_related_field = "username"

    def __str__(self):
        return f"{self.first_name} {self.last_name} ({self.username})"

    class Meta:
        proxy = True
        default_permissions = ("view", "export")
        verbose_name = "Data Query User"
        verbose_name_plural = "Data Query Users"



import os

from django.conf import settings
from django.core.checks import CheckMessage, Warning
from django.core.management import color_style

style = color_style()


def check_etc_dir(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    try:
        settings.ETC_DIR
    except AttributeError:
        pass
    else:
        if not settings.DEBUG and settings.ETC_DIR and not settings.ETC_DIR.startswith("/etc"):
            errors.append(
                Warning(
                    "Insecure configuration. Use root level etc folder. "
                    f"For example, '/etc/{settings.APP_NAME}/' "
                    f"Got {settings.ETC_DIR}",
                    id="settings.ETC_DIR",
                )
            )
        if settings.ETC_DIR and os.access(settings.ETC_DIR, os.W_OK):
            errors.append(
                Warning(
                    "Insecure configuration. Folder is writeable by this user. "
                    f"Got {settings.ETC_DIR}",
                    id="settings.ETC_DIR",
                )
            )
    return errors


def check_static_root(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    try:
        settings.STATIC_ROOT
    except AttributeError:
        pass
    else:
        if settings.STATIC_ROOT and not os.path.exists(settings.STATIC_ROOT):
            errors.append(
                Warning(
                    f"Folder does not exist. Got {settings.STATIC_ROOT}",
                    id="settings.STATIC_ROOT",
                )
            )
    return errors


def check_auth_updater(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    try:
        settings.EDC_AUTH_SKIP_AUTH_UPDATER
    except AttributeError:
        pass
    else:
        if settings.EDC_AUTH_SKIP_AUTH_UPDATER:
            errors.append(
                Warning(
                    "AuthUpdater did not load. Groups and permissions have not been updated. "
                    "See settings.EDC_AUTH_SKIP_AUTH_UPDATER.",
                    id="settings.EDC_AUTH_SKIP_AUTH_UPDATER",
                )
            )
    return errors


def check_site_auths(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    try:
        settings.EDC_AUTH_SKIP_AUTH_UPDATER
    except AttributeError:
        pass
    else:
        if settings.EDC_AUTH_SKIP_AUTH_UPDATER:
            errors.append(
                Warning(
                    "SiteAuths did not autodiscover. Groups and permissions data not ready "
                    "for AuthUpdater. See settings.EDC_AUTH_SKIP_SITE_AUTHS.",
                    id="settings.EDC_AUTH_SKIP_SITE_AUTHS",
                )
            )
    return errors



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_export_admin
from ..models import Plan


@admin.register(Plan, site=edc_export_admin)
class PlanAdmin(admin.ModelAdmin):
    fieldsets = (
        [None, {"fields": ("name", "model", "model_type")}],
        [
            "Export details",
            {
                "fields": (
                    "field_names",
                    "excluded_field_names",
                    "header_labels",
                    "lookups",
                    "header",
                    "encrypt",
                )
            },
        ],
        audit_fieldset_tuple,
    )

    list_display = ("name", "model")
    list_filter = ("name", "model")
    search_fields = ("name", "model")
    radio_fields = {"model_type": admin.VERTICAL}



"""
Based on Corrected Version 2.1 July 2017

Creatinine Clearance14 or eGFR, Low
*Report only one
NA
G1 N/A
G2 < 90 to 60 ml/min or ml/min/1.73 m2 OR 10 to < 30% decrease from participant’s baseline
G3 < 60 to 30 ml/min or ml/min/1.73 m2 OR 30 to < 50% decrease from participant’s baseline
G4 < 30 ml/min or ml/min/1.73 m2 OR ≥ 50% decrease from participant’s baseline
"""

from edc_constants.constants import FEMALE, MALE

from ...adult_age_options import adult_age_options
from ...constants import HIGH_VALUE
from ...formula import Formula
from ...units import (
    CELLS_PER_MILLIMETER_CUBED,
    EGFR_UNITS,
    FEMTOLITERS_PER_CELL,
    GRAMS_PER_DECILITER,
    GRAMS_PER_LITER,
    IU_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
    MILLIMOLES_PER_LITER,
    PERCENT,
    PICOGRAMS_PER_CELL,
    PLUS,
    TEN_X_9_PER_LITER,
)

__all__ = [
    "chemistries",
    "dummies",
    "grading_data",
    "hba1c",
    "hematology",
    "urinalysis",
]

dummies = {
    "hba1c": [
        Formula(
            "x<0",
            grade=0,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "hct": [
        Formula(
            "x<0",
            grade=0,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "hdl": [
        Formula(
            "x<0",
            grade=0,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "ggt": [
        Formula(
            "x<0",
            grade=0,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "mch": [
        Formula(
            "x<0",
            grade=0,
            units=PICOGRAMS_PER_CELL,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "mchc": [
        Formula(
            "x<0",
            grade=0,
            units=GRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "mcv": [
        Formula(
            "x<0",
            grade=0,
            units=FEMTOLITERS_PER_CELL,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "rbc": [
        Formula(
            "x<0",
            grade=0,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<0",
            grade=0,
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "urea": [
        Formula(
            "x<0",
            grade=0,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "crp": [
        Formula(
            "x<0",
            grade=0,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<0",
            grade=0,
            units=MILLIGRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
}


chemistries = dict(
    albumin=[
        Formula(
            "x<2.0",
            grade=3,
            units=GRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<20",
            grade=3,
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    alp=[
        Formula(
            "1.25*ULN<=x<2.50*ULN",
            grade=1,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.50*ULN<=x<5.00*ULN",
            grade=2,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x<10.00*ULN",
            grade=3,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "10.00*ULN<=x",
            grade=4,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    alt=[
        Formula(
            "1.25*ULN<=x<2.50*ULN",
            grade=1,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.50*ULN<=x<5.00*ULN",
            grade=2,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x<10.00*ULN",
            grade=3,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "10.00*ULN<=x",
            grade=4,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    amylase=[
        Formula(
            "1.1*ULN<=x<1.5*ULN",
            grade=1,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.5*ULN<=x<3.0*ULN",
            grade=2,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "3.0*ULN<=x<5.0*ULN",
            grade=3,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            f"5.0*ULN<=x<{HIGH_VALUE}*ULN",
            grade=4,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    ast=[
        Formula(
            "1.25*ULN<=x<2.50*ULN",
            grade=1,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.50*ULN<=x<5.00*ULN",
            grade=2,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x<10.00*ULN",
            grade=3,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "10.00*ULN<=x",
            grade=4,
            units=IU_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    chol=[
        Formula(
            "300<=x",
            grade=3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "7.77<=x",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    creatinine=[
        Formula(
            "1.1*ULN<=x<=1.3*ULN",
            grade=1,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.3*ULN<x<=1.8*ULN",
            grade=2,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.8*ULN<x<3.5*ULN",
            grade=3,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "3.5*ULN<=x",
            grade=4,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.1*ULN<=x<=1.3*ULN",
            grade=1,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.3*ULN<x<=1.8*ULN",
            grade=2,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.8*ULN<x<3.5*ULN",
            grade=3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "3.5*ULN<=x",
            grade=4,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    egfr=[  # not considering % drop
        Formula(
            "60<=x<90",
            grade=2,
            units=EGFR_UNITS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "30<=x<60",
            grade=3,
            units=EGFR_UNITS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<30",
            grade=4,
            units=EGFR_UNITS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    egfr_drop=[  # % drop from baseline
        Formula(
            "10<=x<30",
            grade=2,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "30<=x<50",
            grade=3,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "50<=x",
            grade=4,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    # glucose=[  # G3/G4 same for fasting / non-fasting
    #     Formula(
    #         "13.89<=x<27.75",
    #         grade=3,
    #         units=MILLIMOLES_PER_LITER,
    #         gender=[MALE, FEMALE],
    #         **adult_age_options,
    #         fasting=True,
    #     ),
    #     Formula(
    #         "27.75<=x",
    #         grade=4,
    #         units=MILLIMOLES_PER_LITER,
    #         gender=[MALE, FEMALE],
    #         **adult_age_options,
    #         fasting=True,
    #     ),
    #     Formula(
    #         "13.89<=x<27.75",
    #         grade=3,
    #         units=MILLIMOLES_PER_LITER,
    #         gender=[MALE, FEMALE],
    #         **adult_age_options,
    #         fasting=False,
    #     ),
    #     Formula(
    #         "27.75<=x",
    #         grade=4,
    #         units=MILLIMOLES_PER_LITER,
    #         gender=[MALE, FEMALE],
    #         **adult_age_options,
    #         fasting=False,
    #     ),
    # ],
    ldl=[
        Formula(
            "4.90<=x",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
            fasting=True,
        ),
    ],
    magnesium=[
        Formula(
            "0.30<=x<=0.44",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<0.30",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.7<=x<=1.1",
            grade=3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<0.7",
            grade=4,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    potassium=[
        Formula(
            "2.0<=x<2.5",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<2.0",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "6.5<=x<7.0",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "7.0<=x",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    sodium=[
        Formula(
            "121<=x<125",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "154<=x<160",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "160<=x",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<=120",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    tbil=[
        Formula(
            "1.10*ULN<=x<1.60*ULN",
            grade=1,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.60*ULN<=x<2.60*ULN",
            grade=2,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.60*ULN<=x<5.00*ULN",
            grade=3,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x",
            grade=4,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.10*ULN<=x<1.60*ULN",
            grade=1,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.60*ULN<=x<2.60*ULN",
            grade=2,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.60*ULN<=x<5.00*ULN",
            grade=3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x",
            grade=4,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.10*ULN<=x<1.60*ULN",
            grade=1,
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.60*ULN<=x<2.60*ULN",
            grade=2,
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2.60*ULN<=x<5.00*ULN",
            grade=3,
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.00*ULN<=x",
            grade=4,
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    trig=[
        Formula(
            "5.7<=x<=11.4",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
            fasting=True,
        ),
        Formula(
            "11.4<x",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
            fasting=True,
        ),
    ],
    uric_acid=[
        Formula(
            "12.0<=x<15.0",
            grade=3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "15.0<=x",
            grade=4,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.71<=x<0.89",
            grade=3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.89<=x",
            grade=4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
)

hematology = {
    "haemoglobin": [
        Formula(
            "7.0<=x<9.0",
            grade=3,
            units=GRAMS_PER_DECILITER,
            gender=[MALE],
            **adult_age_options,
        ),
        Formula(
            "6.5<=x<8.5",
            grade=3,
            units=GRAMS_PER_DECILITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<7.0",
            grade=4,
            units=GRAMS_PER_DECILITER,
            gender=[MALE],
            **adult_age_options,
        ),
        Formula(
            "x<6.5",
            grade=4,
            units=GRAMS_PER_DECILITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
    ],
    "platelets": [
        Formula(
            "25<=x<=50",
            grade=3,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<25",
            grade=4,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "25000<=x<=50000",
            grade=3,
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<25000",
            grade=4,
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "neutrophil": [
        Formula(
            "0.40<=x<=0.59",
            grade=3,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<0.40",
            grade=4,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "wbc": [
        Formula(
            "1.00<=x<=1.49",
            grade=3,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "x<1.00",
            grade=4,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
}

urinalysis = {
    "proteinuria": [
        Formula(
            "1<=x<2",
            grade=1,
            units=PLUS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "2<=x<3",
            grade=2,
            units=PLUS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "3<=x",
            grade=3,
            units=PLUS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
}

hba1c = {
    "hba1c": [
        Formula(
            "9999999<=x<=99999999",
            grade=3,
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ]
}

grading_data = {}
grading_data.update(**dummies)
grading_data.update(**chemistries)
grading_data.update(**hematology)
grading_data.update(**hba1c)
grading_data.update(**urinalysis)



from django import forms

from ...models import Dispense


class DispenseForm(forms.ModelForm):
    class Meta:
        model = Dispense
        fields = "__all__"
        help_text = {"dispense_identifier": "(read-only)"}
        widgets = {
            "dispense_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from .confirmation_form import ConfirmationForm
from .container_form import ContainerForm
from .container_type_form import ContainerTypeForm
from .container_units_form import ContainerUnitsForm
from .dispense_form import DispenseForm
from .location_form import LocationForm
from .lot_form import LotForm
from .order_form import OrderForm, OrderFormSuper
from .order_item_form import OrderItemForm
from .product_form import ProductForm
from .receive_form import ReceiveForm, ReceiveFormSuper
from .receive_item_form import ReceiveItemForm
from .repack_request_form import RepackRequestForm
from .stock_form import StockForm
from .stock_request_form import StockRequestForm
from .stock_request_item_form import StockRequestItemForm
from .stock_transfer_form import StockTransferForm
from .storage_bin_form import StorageBinForm
from .storage_bin_item_form import StorageBinItemForm
from .supplier_form import SupplierForm



from django.apps import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_lab"
    verbose_name = "Edc Lab"
    has_exportable_data = True
    include_in_administration_section = True
    result_model = "edc_lab.result"



# taken from django-extensions management command
# https://github.com/django-extensions/django-extensions/blob/master/django_extensions/management/commands/show_urls.py

from django.core.exceptions import ViewDoesNotExist
from django.urls import URLPattern, URLResolver, get_resolver


class RegexURLPattern:
    pass


class RegexURLResolver:
    pass


class LocaleRegexURLResolver:
    pass


def describe_pattern(p):
    return str(p.pattern)


def show_namespaces():
    """Returns a list of all registered URL namespaces."""
    resolver = get_resolver()
    return list(resolver.namespace_dict.keys())


def show_urls():
    """Return a list of all URL patterns in the project."""
    urls = []
    resolver = get_resolver()
    url_patterns = resolver.url_patterns
    for pattern in url_patterns:
        try:
            urls.append(f"URL: {pattern.pattern} -> View: {pattern.callback.__name__}")
        except AttributeError:
            urls.append(f"URL: {pattern.pattern}")
    return urls


def show_urls_from_patterns(urlpatterns, base="", namespace=None, search=None):
    urls = extract_views_from_urlpatterns(urlpatterns, base=base, namespace=namespace)
    if search:
        return [url[1] for url in urls if search in url[1]]
    return [url[1] for url in urls]


def show_url_names(urlpatterns, base="", namespace=None, search=None):
    """Returns a list of url names.

    For example:
        resolver = get_resolver()
        url_patterns = resolver.url_patterns
        pprint(show_url_names(url_patterns))
    """

    urls = extract_views_from_urlpatterns(urlpatterns, base=base, namespace=namespace)
    if search:
        return [url[2] for url in urls if search in url[2]]
    return [url[2] for url in urls]


def extract_views_from_urlpatterns(urlpatterns, base="", namespace=None):  # noqa
    """
    Return a list of views from a list of urlpatterns.
    Each object in the returned list is a three-tuple: (view_func, regex, name)
    """
    views = []
    for p in urlpatterns:
        if isinstance(p, (URLPattern, RegexURLPattern)):
            try:
                if not p.name:
                    name = p.name
                elif namespace:
                    name = f"{namespace}:{p.name}"
                else:
                    name = p.name
                pattern = describe_pattern(p)
                views.append((p.callback, base + pattern, name))
            except ViewDoesNotExist:
                continue
        elif isinstance(p, (URLResolver, RegexURLResolver)):
            try:
                patterns = p.url_patterns
            except ImportError:
                continue
            if namespace and p.namespace:
                _namespace = f"{namespace}:{p.namespace}"
            else:
                _namespace = p.namespace or namespace
            pattern = describe_pattern(p)
            views.extend(
                extract_views_from_urlpatterns(patterns, base + pattern, namespace=_namespace)
            )
        elif hasattr(p, "_get_callback"):
            try:
                views.append((p._get_callback(), base + describe_pattern(p), p.name))
            except ViewDoesNotExist:
                continue
        elif hasattr(p, "url_patterns") or hasattr(p, "_get_url_patterns"):
            try:
                patterns = p.url_patterns
            except ImportError:
                continue
            views.extend(
                extract_views_from_urlpatterns(
                    patterns, base + describe_pattern(p), namespace=namespace
                )
            )
        else:
            raise TypeError(f"{p} does not appear to be a urlpattern object")
    return views



# Generated by Django 2.2.2 on 2019-09-22 02:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_offstudy", "0005_auto_20190305_0123")]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectoffstudy",
            name="subject_identifier",
            field=models.CharField(db_index=True, max_length=50),
        ),
        migrations.AlterField(
            model_name="subjectoffstudy",
            name="subject_identifier",
            field=models.CharField(max_length=50, unique=True),
        ),
    ]



# Generated by Django 3.1.8 on 2021-04-25 13:28

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_adverse_event", "0005_auto_20210120_0005"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="aeclassification",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ["display_index", "display_name"],
            },
        ),
        migrations.AlterModelOptions(
            name="causeofdeath",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ["display_index", "display_name"],
            },
        ),
    ]



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from edc_action_item.fieldsets import action_fields, action_fieldset_tuple
from edc_action_item.modeladmin_mixins import ActionItemModelAdminMixin
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin


class HospitalizationModelAdminMixin(
    ModelAdminSubjectDashboardMixin,
    ActionItemModelAdminMixin,
):

    add_form_template: str = "edc_adverse_event/admin/change_form.html"
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"

    fieldsets = (
        (None, {"fields": ("subject_identifier", "report_datetime")}),
        (
            "Hospital admission",
            {
                "fields": (
                    "have_details",
                    "admitted_date",
                    "admitted_date_estimated",
                )
            },
        ),
        (
            "Hospital discharge",
            {
                "fields": (
                    "discharged",
                    "discharged_date",
                    "discharged_date_estimated",
                )
            },
        ),
        ("Narrative", {"fields": ("narrative",)}),
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    list_display = ("subject_identifier", "action_identifier")

    radio_fields = {
        "admitted_date_estimated": admin.VERTICAL,
        "discharged": admin.VERTICAL,
        "discharged_date_estimated": admin.VERTICAL,
        "have_details": admin.VERTICAL,
    }

    search_fields = ("subject_identifier", "action_identifier")

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        action_flds = tuple(fld for fld in action_fields if fld != "action_identifier")
        return tuple(set(action_flds + readonly_fields))



from __future__ import annotations

from datetime import date

from django import forms

from edc_utils.date import to_utc
from edc_utils.text import formatted_datetime

from .base_form_validator import BaseFormValidator


def convert_any_to_date(date1, date2) -> tuple[date, date]:
    try:
        date1 = date1.date()
    except AttributeError:
        pass
    try:
        date2 = date2.date()
    except AttributeError:
        pass
    return date1, date2


class DateRangeFieldValidator(BaseFormValidator):
    def date_not_before(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        convert_to_date: bool | None = None,
        message_on_field=None,
    ) -> None:
        """Asserts date_field2 is not before date_field1"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        msg = msg or f"Invalid. Cannot be before {date_field1} "
        if date1 and date2:
            if date1 > date2:
                raise forms.ValidationError({message_on_field or date_field2: f"{msg}."})

    def date_not_after(
        self,
        date_field1: str,
        date_field2: str,
        msg: str | None = None,
        convert_to_date: bool | None = None,
    ) -> None:
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        msg = msg or f"Invalid. Cannot be after {date_field1} "
        if date1 and date2:
            if date1 < date2:
                raise forms.ValidationError({date_field2: f"{msg}"})

    def date_equal(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        message_on_field=None,
        convert_to_date: bool | None = None,
    ) -> None:
        """Asserts date1 and date2 are equal"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        msg = msg or f"Invalid. Expected {date_field2} to be the same as {date_field1}."
        if date1 and date2:
            if date1 != date2:
                raise forms.ValidationError({message_on_field or date_field2: f"{msg}"})

    def date_not_equal(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        message_on_field=None,
        convert_to_date: bool | None = None,
    ) -> None:
        """Asserts date1 and date2 are equal"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        msg = msg or f"Invalid. Expected {date_field2} to be the same as {date_field1}."
        if date1 and date2:
            if date1 == date2:
                raise forms.ValidationError({message_on_field or date_field2: f"{msg}"})

    def datetime_not_before(
        self, datetime_field1: str, datetime_field2: str, msg=None
    ) -> None:
        datetime1 = self.cleaned_data.get(datetime_field1)
        datetime2 = self.cleaned_data.get(datetime_field2)
        if datetime1:
            datetime1 = to_utc(datetime1)
        if datetime2:
            datetime2 = to_utc(datetime2)
        msg = msg or f"Invalid. Cannot be before {datetime_field2} "
        if datetime1 and datetime2:
            if datetime1 < datetime2:
                raise forms.ValidationError(
                    {datetime_field1: f"{msg}. Got {formatted_datetime(datetime2)}."}
                )

    def datetime_not_after(self, datetime_field1: str, datetime_field2: str, msg=None) -> None:
        datetime_field1 = self.cleaned_data.get(datetime_field1)
        datetime_field2 = self.cleaned_data.get(datetime_field2)
        msg = msg or f"Invalid. Cannot be before date of {datetime_field2} "
        if datetime_field1:
            datetime_field1 = to_utc(datetime_field1)
            datetime_field2 = to_utc(datetime_field2)
            if datetime_field1 > datetime_field2:
                raise forms.ValidationError({datetime_field1: f"{msg}"})

    def datetime_equal(self, datetime_field1: str, datetime_field2: str, msg=None) -> None:
        datetime_field1 = self.cleaned_data.get(datetime_field1)
        datetime_field2 = self.cleaned_data.get(datetime_field2)
        msg = msg or f"Invalid. Cannot be before date of {datetime_field2} "
        if datetime_field1:
            datetime_field1 = to_utc(datetime_field1)
            datetime_field2 = to_utc(datetime_field2)
            if datetime_field1 == datetime_field2:
                raise forms.ValidationError(
                    {datetime_field1: f"{msg}. Got {formatted_datetime(datetime_field2)}."}
                )



from edc_auth.constants import (
    CELERY_MANAGER,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
)
from edc_auth.site_auths import site_auths

from .auth_objects import (
    DATA_MANAGER,
    DATA_MANAGER_EXPORT,
    DATA_MANAGER_ROLE,
    DATA_MANAGER_SUPER,
    DATA_QUERY,
    DATA_QUERY_VIEW,
    SITE_DATA_MANAGER_ROLE,
    custom_codename_tuples,
    data_manager,
)


def update_site_auths():
    site_auths.add_custom_permissions_tuples(
        model="edc_data_manager.edcpermissions", codename_tuples=custom_codename_tuples
    )

    site_auths.add_custom_permissions_tuples(
        model="edc_data_manager.edcpermissions",
        codename_tuples=[("edc_data_manager.special_bypassmodelform", "Can bypass modelform")],
    )

    # groups
    site_auths.add_group(*data_manager, name=DATA_MANAGER)
    site_auths.add_group(*data_manager, name=DATA_QUERY_VIEW, view_only=True)

    site_auths.add_group(*data_manager, name=DATA_QUERY, view_only=True)
    site_auths.update_group("edc_data_manager.change_dataquery", name=DATA_QUERY)

    site_auths.add_group(
        "edc_data_manager.export_datadictionary",
        "edc_data_manager.export_dataquery",
        "edc_data_manager.export_queryrule",
        name=DATA_MANAGER_EXPORT,
    )

    site_auths.add_group(*data_manager, name=DATA_MANAGER_SUPER)
    site_auths.update_group("edc_data_manager.change_dataquery", name=DATA_MANAGER_SUPER)
    site_auths.update_group(
        "edc_data_manager.export_datadictionary",
        "edc_data_manager.export_dataquery",
        "edc_data_manager.export_queryrule",
        "edc_data_manager.special_bypassmodelform",
        "edc_data_manager.change_dataquery",
        name=DATA_MANAGER_SUPER,
    )

    # roles
    site_auths.add_role(CELERY_MANAGER, DATA_MANAGER, name=DATA_MANAGER_ROLE)
    site_auths.add_role(DATA_QUERY, name=SITE_DATA_MANAGER_ROLE)
    site_auths.update_role(DATA_QUERY, name=CLINICIAN_ROLE)
    site_auths.update_role(DATA_QUERY, name=NURSE_ROLE)
    site_auths.update_role(DATA_QUERY, name=CLINICIAN_SUPER_ROLE)


update_site_auths()



from ..metadata import CrfMetadataGetter
from .crf_metadata_wrapper import CrfMetadataWrapper
from .metadata_wrappers import MetadataWrappers


class CrfMetadataWrappers(MetadataWrappers):
    metadata_getter_cls = CrfMetadataGetter
    metadata_wrapper_cls = CrfMetadataWrapper



import sys
from pathlib import Path

from django.core.management.base import BaseCommand
from django.core.management.color import color_style
from django_pandas.io import read_frame

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_reportable.models import GradingData, NormalData

style = color_style()


def export_daids_grading(path: str, reference_range_collection_name: str | None = None):
    path = Path(path or "~/").expanduser()
    sys.stdout.write(style.MIGRATE_HEADING("Exporting reportables to document (.csv) ...\n"))
    reference_range_collection_name = (
        reference_range_collection_name or ResearchProtocolConfig().project_name.lower()
    )
    df = read_frame(
        GradingData.objects.filter(
            reference_range_collection__name=reference_range_collection_name
        ),
        verbose=True,
    )
    fname = path / f"{reference_range_collection_name}_grading_data.csv"
    df.to_csv(fname, index=False)
    sys.stdout.write(style.MIGRATE_HEADING(f"  * Exported grading data to `{fname}`\n"))
    df = read_frame(
        NormalData.objects.filter(
            reference_range_collection__name=reference_range_collection_name
        ),
        verbose=True,
    )
    fname = path / f"{reference_range_collection_name}_normal_data.csv"
    df.to_csv(fname, index=False)
    sys.stdout.write(style.MIGRATE_HEADING(f"  * Exported normal data to `{fname}`\n"))
    sys.stdout.write(style.MIGRATE_HEADING("Done\n"))


class Command(BaseCommand):
    help = "Export export DAIDS grading to document (.csv)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--path",
            default="~/",
            action="store_true",
            dest="path",
            help="Export path/folder",
        )

    def handle(self, *args, **options):
        export_daids_grading(path=options["path"])



import pandas as pd

from ..constants import SYSTEM_COLUMNS
from ..dialects import RsDialect
from .df_handler import DfHandler


class RegisteredSubjectDfHandler(DfHandler):
    rs_dialect_cls = RsDialect

    registered_subject_tbl = "edc_registration_registeredsubject"
    registered_subject_column = "registered_subject_id"
    system_columns = SYSTEM_COLUMNS
    sort_by = ["subject_identifier"]
    exclude_export_columns = True
    exclude_system_columns = False

    def __init__(self, exclude_system_columns=None, **kwargs):
        self._df_registered_subject = pd.DataFrame()
        self.rs_dialect = self.rs_dialect_cls(self)
        self.exclude_system_columns = exclude_system_columns or self.exclude_system_columns
        super().__init__(**kwargs)

    def prepare_dataframe(self, **kwargs):
        """Merges CRF dataframe with df_visit_and_related
        on visit_column.
        """
        self.dataframe = pd.merge(
            left=self.dataframe,
            right=self.df_visit_and_related,
            how="left",
            on=self.visit_column,
            suffixes=["_notused", ""],
        )
        self.dataframe = self.dataframe[self.columns]

    @property
    def columns(self):
        """Returns a list of column names."""
        df_columns = list(self.dataframe.columns)
        df_columns.pop(df_columns.index(self.registered_subject_column))
        columns = list(self.df_registered_subject.columns)
        columns.extend([c for c in df_columns if c not in columns])
        # "export_" columns
        if self.exclude_export_columns:
            columns = [col for col in columns if not col.startswith("export_")]
        # "system" columns, move to the end
        if not self.exclude_system_columns:
            columns = [col for col in columns if col not in self.system_columns]
            columns.extend(self.system_columns)
        return columns

    @property
    def df_registered_subject(self):
        """Returns a dataframe of the crf_dialect's `select_visit_and_related`
        SQL statement.
        """
        if self._df_registered_subject.empty:
            sql, params = self.rs_dialect_cls.select_registered_subject
            self._df_registered_subject = self.db.read_sql(sql, params=params)
        return self._df_registered_subject



# Generated by Django 5.1.2 on 2024-11-04 00:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pylabels", "0005_labelconfiguration_delete_label_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Drawing",
        ),
    ]



from django import forms

from ...models import DosageGuideline


class DosageGuidelineForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("dose") and cleaned_data.get("dose_per_kg"):
            raise forms.ValidationError("Specify either dose or dose per kg, not both")
        if not cleaned_data.get("dose") and not cleaned_data.get("dose_per_kg"):
            raise forms.ValidationError("Either dose or dose per kg is required.")
        return cleaned_data

    class Meta:
        model = DosageGuideline
        fields = "__all__"



from typing import TYPE_CHECKING

from celery.states import PENDING
from django.contrib import admin, messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext

from edc_utils.celery import get_task_result

if TYPE_CHECKING:

    from ...models import StockRequest


@admin.display(description="Prepare stock request items")
def prepare_stock_request_items_action(modeladmin, request, queryset):
    """
    1. is there an open unprocess stock request?
    2. what stock is available at the site?
    3. what stock is available at central?

    """
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        stock_request: StockRequest = queryset.first()
        if getattr(get_task_result(stock_request), "status", "") == PENDING:
            messages.add_message(
                request,
                messages.ERROR,
                (
                    f"Stock request {stock_request.request_identifier} is still processing. "
                    "Please click cancel and check the status column."
                ),
            )
        else:
            url = reverse(
                "edc_pharmacy:review_stock_request_url",
                kwargs={"stock_request": stock_request.pk},
            )
            return HttpResponseRedirect(url)
    return None



import hashlib

import numpy as np
import pandas as pd

from .decrypt import decrypt


def identity(row, column_name=None):
    return decrypt(row, column_name or "identity", "rsa", "restricted")


def identity256_decrypt(row, column_name=None):
    identity256 = np.nan
    column_name = column_name or "identity"
    if pd.notnull(row[column_name]):
        identity256 = hashlib.sha256(identity(row, column_name)).digest().encode("hex")
    return identity256


def identity256(row, column_name=None):
    identity256 = np.nan
    column_name = column_name or "identity"
    if pd.notnull(row[column_name]):
        identity256 = hashlib.sha256(row[column_name].encode("utf-8")).hexdigest()
    return identity256



from django.apps import apps as django_apps
from django.core import serializers
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.utils import timezone

from edc_constants.constants import NEW

from ..constants import EXPORTED, INSERT, UPDATE


class ObjectHistoryUpdaterError(Exception):
    pass


class Base:
    history_model = "edc_export.objecthistory"

    @property
    def model_cls(self):
        return django_apps.get_model(self.history_model)


class ObjectHistoryCreator(Base):
    def create(self, model_obj=None, change_type=None, using=None):
        if not change_type:
            change_type = self.get_change_type(model_obj=model_obj)
        export_datetime = timezone.now()
        if model_obj._meta.proxy_for_model:  # if proxy model, get main model
            model_obj = model_obj._meta.proxy_for_model.objects.get(id=model_obj.id)
        obj = self.model_cls.objects.using(using).create(
            model=model_obj._meta.label_lower,
            tx_pk=model_obj.id,
            export_change_type=change_type,
            exported=False,
            export_uuid=model_obj.export_uuid,
            status=NEW,
            tx=self.get_json_tx(model_obj),
            exported_datetime=export_datetime,
            timestamp=export_datetime.strftime("%Y%m%d%H%M%S%f"),
        )
        return obj

    def get_json_tx(self, model_obj=None):
        return serializers.serialize(
            "json",
            [model_obj],
            ensure_ascii=True,
            use_natural_foreign_keys=True,
            use_natural_primary_keys=False,
        )

    def get_change_type(self, model_obj=None):
        """Returns the export_change_type by querying the
        object history model for existing instances.
        """
        try:
            self.model_cls.objects.get(export_uuid=model_obj.export_uuid)
        except ObjectDoesNotExist:
            export_change_type = INSERT
        except MultipleObjectsReturned:
            export_change_type = UPDATE
        else:
            export_change_type = UPDATE
        return export_change_type


class ObjectHistoryGetter(Base):
    object_creator = ObjectHistoryCreator()

    def get_not_exported(self, model_obj=None, create=None):
        """Returns a queryset of objecthistory objects
        not yet exported.
        """
        try:
            self.model_cls.objects.get(export_uuid=model_obj.export_uuid)
        except ObjectDoesNotExist:
            if create:
                self.object_creator.create(model_obj=model_obj)
        except MultipleObjectsReturned:
            pass
        return self.model_cls.objects.filter(
            export_uuid=model_obj.export_uuid, exported=False
        ).order_by("created")


class ObjectHistoryUpdater(Base):
    def update_as_exported(self, objects=None, exported_datetime=None):
        """Updates all objects in the objecthistory queryset
        as exported.
        """
        for obj in objects:
            if obj.exported:
                raise ObjectHistoryUpdaterError(f"Already exported. Got {objects}.")
        objects.update(
            exported=True,
            status=EXPORTED,
            exported_datetime=exported_datetime,
            timestamp=exported_datetime.strftime("%Y%m%d%H%M%S"),
        )


class ObjectHistoryHelper:
    obj_getter = ObjectHistoryGetter()
    obj_updater = ObjectHistoryUpdater()

    def __init__(self, model_obj=None, create=None):
        self.model_obj = model_obj
        self.create = create

    def get_not_exported(self):
        return self.obj_getter.get_not_exported(model_obj=self.model_obj, create=self.create)

    def update_as_exported(self, **kwargs):
        self.obj_updater.update_as_exported(**kwargs)



from __future__ import annotations

from tempfile import mkdtemp
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from reportlab.graphics.barcode import code39
from reportlab.lib import colors
from reportlab.lib.units import cm, mm
from reportlab.platypus import Paragraph, Spacer, Table, TableStyle

from edc_pdf_reports import Report

from ..models import Aliquot, Box, BoxItem
from ..site_labs import site_labs

if TYPE_CHECKING:
    from django.contrib.auth.models import User

    from edc_model.models import BaseUuidModel
    from edc_pdf_reports.model_mixins import PdfReportModelMixin

    from ..model_mixins import ManifestModelMixin

    class Model(PdfReportModelMixin, ManifestModelMixin, BaseUuidModel): ...


class ManifestPdfReportError(Exception):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code


class ManifestPdfReport(Report):
    def __init__(self, model_obj: Model = None, user: User = None, **kwargs):
        super().__init__(**kwargs)
        self.manifest = model_obj
        self.user = user
        self.box_model = Box
        self.box_item_model = BoxItem
        self.aliquot_model = Aliquot
        self.image_folder = mkdtemp()
        if self.manifest.shipped:
            self.filename = f"{self.manifest.manifest_identifier}.pdf"
        else:
            self.filename = "manifest_preview.pdf"

    def get_report_story(self, **kwargs):
        story = []

        data = [
            [
                Paragraph(
                    " ".join(self.manifest.site.name.split("_")).upper(),
                    self.styles["line_data_large"],
                )
            ],
            [Paragraph("SITE NAME", self.styles["line_label"])],
            [
                Paragraph(
                    self.formatted_address(**self.manifest.shipper.__dict__),
                    self.styles["line_data_large"],
                )
            ],
            [Paragraph("SITE DETAILS", self.styles["line_label"])],
        ]

        t = Table(data, colWidths=(9 * cm))
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (0, 3), 0.25, colors.black),
                ]
            )
        )
        t.hAlign = "RIGHT"

        story.append(t)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Paragraph(
                "SPECIMEN MANIFEST{}".format(" (reprint)" if self.manifest.printed else ""),
                self.styles["line_label_center"],
            )
        )

        if self.manifest.shipped:
            barcode = code39.Standard39(
                self.manifest.manifest_identifier, barHeight=10 * mm, stop=1
            )
        else:
            barcode = "PREVIEW"

        data = [
            [
                Paragraph("MANIFEST NO.", self.styles["line_label"]),
                Paragraph(
                    (
                        self.manifest.human_readable_identifier
                        if self.manifest.shipped
                        else "PREVIEW"
                    ),
                    self.styles["line_data_largest"],
                ),
                barcode,
            ]
        ]
        t = Table(data, colWidths=(3 * cm, None, 5.5 * cm))
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (1, 0), (-1, -1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )

        story.append(t)

        data = [
            [
                Paragraph("MANIFEST DATE", self.styles["line_label"]),
                Paragraph("EXPORT REFERENCES", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.manifest.manifest_datetime.strftime("%Y-%m-%d"),
                    self.styles["line_data_largest"],
                ),
                Paragraph(
                    self.manifest.export_references or "",
                    self.styles["line_data_largest"],
                ),
            ],
        ]
        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph(
                    "SHIPPER/EXPORTER (complete name and address)",
                    self.styles["line_label"],
                ),
                Paragraph("CONSIGNEE (complete name and address)", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.formatted_address(**self.shipper_data),
                    self.styles["line_data_large"],
                ),
                Paragraph(
                    self.formatted_address(**self.consignee_data),
                    self.styles["line_data_large"],
                ),
            ],
        ]

        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data1 = [
            [
                Paragraph("COUNTRY OF EXPORT", self.styles["line_label"]),
                Paragraph(
                    "DESCRIPTION OF GOODS (description and special instructions)<br />",
                    self.styles["line_label"],
                ),
            ],
            [
                Paragraph(self.manifest.shipper.country, self.styles["line_data_largest"]),
                Paragraph(self.description, self.styles["line_data_large"]),
            ],
            [Paragraph("COUNTRY OF ORIGIN", self.styles["line_label"]), ""],
            [
                Paragraph(self.manifest.shipper.country, self.styles["line_data_largest"]),
                "",
            ],
            [
                Paragraph("COUNTRY OF ULTIMATE DESTINATION", self.styles["line_label"]),
                "",
            ],
            [
                Paragraph(self.manifest.consignee.country, self.styles["line_data_largest"]),
                "",
            ],
        ]
        t1 = Table(data1)
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 1), (0, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (0, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (1, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (1, 3), 0.25, colors.black),
                    ("INNERGRID", (0, 4), (1, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 5), (1, 5), 0.25, colors.black),
                    ("SPAN", (1, 1), (1, 5)),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t1)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Table(
                [
                    [
                        Paragraph(
                            "I DECLARE THE INFORMATION CONTAINED IN THIS "
                            "MANIFEST TO BE TRUE AND CORRECT",
                            self.styles["line_label"],
                        )
                    ]
                ]
            )
        )

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        data1 = [
            [
                Paragraph(self.contact_name, self.styles["line_data_large"]),
                "",
                Paragraph(
                    (
                        self.manifest.export_datetime.strftime("%Y-%m-%d %H:%M")
                        if self.manifest.shipped
                        else "PREVIEW"
                    ),
                    self.styles["line_data_large"],
                ),
            ],
            [
                Paragraph(
                    "SIGNATURE OF SHIPPER/EXPORTER (Type name and title and sign.)",
                    self.styles["line_label"],
                ),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data1, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )

        story.append(t1)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        # boxes
        story.append(
            Table([[Paragraph("MANIFEST CONTENTS", self.styles["line_label_center"])]])
        )

        story = self.append_manifest_items_story(story)

        if self.manifest.shipped and not self.manifest.printed:
            self.manifest.printed = True
            self.manifest.save()

        return story

    @property
    def contact_name(self):
        return f"{self.user.first_name} {self.user.last_name}"

    @property
    def shipper_data(self):
        data = self.manifest.shipper.__dict__
        if self.contact_name.strip():
            data.update(contact_name=self.contact_name)
        return data

    @property
    def consignee_data(self):
        data = self.manifest.consignee.__dict__
        return data

    def formatted_address(self, **kwargs):
        data = {
            "contact_name": None,
            "name": None,
            "address": None,
            "city": None,
            "state": None,
            "postal_code": "0000",
            "country": None,
        }
        data.update(**kwargs)
        data_list = [
            v
            for v in [
                data.get("contact_name"),
                data.get("name"),
                data.get("address"),
                (
                    data.get("city")
                    if not data.get("state")
                    else "{} {}".format(data.get("city"), data.get("state"))
                ),
                data.get("postal_code"),
                data.get("country"),
            ]
            if v
        ]
        return "<br />".join(data_list)

    @property
    def description(self):
        boxes = self.box_model.objects.filter(
            box_identifier__in=[obj.identifier for obj in self.manifest.manifestitem_set.all()]
        )
        box_items = self.box_item_model.objects.filter(box__in=boxes)
        aliquots = self.aliquot_model.objects.filter(
            aliquot_identifier__in=[obj.identifier for obj in box_items]
        )
        specimen_types = list(set([obj.aliquot_type for obj in aliquots]))
        description = {
            "box_count": boxes.count(),
            "specimen_count": aliquots.count(),
            "specimen_types": ", ".join(specimen_types),
        }
        box_word = "box" if boxes.count() == 1 else "boxes"
        specimen_word = "specimen" if aliquots.count() == 1 else "specimens"
        type_word = "type" if len(specimen_types) == 1 else "types"
        return (
            "{box_count} {box_word} containing {specimen_count} "
            "{specimen_word} of {type_word} {specimen_types}.".format(
                box_word=box_word,
                specimen_word=specimen_word,
                type_word=type_word,
                **description,
            )
        )

    def append_manifest_items_story(self, story):
        box_header = [
            Paragraph("BOX:", self.styles["line_label"]),
            Paragraph("CATEGORY:", self.styles["line_label"]),
            Paragraph("TYPES:", self.styles["line_label"]),
            Paragraph("ITEMS:", self.styles["line_label"]),
            Paragraph("BOX DATE:", self.styles["line_label"]),
            Paragraph("BOX BARCODE:", self.styles["line_label"]),
        ]
        for index, manifest_item in enumerate(
            self.manifest.manifestitem_set.all().order_by("-created")
        ):
            if index > 0:
                story.append(Spacer(0.1 * cm, 0.5 * cm))
            data1 = []
            data1.append(box_header)
            try:
                box = self.box_model.objects.get(box_identifier=manifest_item.identifier)
            except ObjectDoesNotExist as e:
                raise ManifestPdfReportError(
                    f"{e} Got Manifest item '{manifest_item.identifier}'.",
                    code="unboxed_item",
                ) from e
            barcode = code39.Standard39(box.box_identifier, barHeight=5 * mm, stop=1)
            data1.append(
                [
                    Paragraph(box.box_identifier, self.styles["line_data_large"]),
                    Paragraph(
                        box.get_category_display().upper(),
                        self.styles["line_data_large"],
                    ),
                    Paragraph(box.specimen_types, self.styles["line_data_large"]),
                    Paragraph(
                        f"{box.count!s}/{box.box_type.total!s}",
                        self.styles["line_data_large"],
                    ),
                    Paragraph(
                        box.box_datetime.strftime("%Y-%m-%d"),
                        self.styles["line_data_large"],
                    ),
                    barcode,
                ]
            )
            t1 = Table(data1, colWidths=(None, None, None, None, None, None))
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, 0), 0.25, colors.black),
                        ("INNERGRID", (0, 1), (-1, -1), 0.25, colors.black),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)
            story.append(Spacer(0.1 * cm, 0.5 * cm))

            table_data = []
            table_data = [
                [
                    Paragraph("BARCODE", self.styles["line_label_center"]),
                    Paragraph("POS", self.styles["line_label_center"]),
                    Paragraph("ALIQUOT IDENTIFIER", self.styles["line_label_center"]),
                    Paragraph("SUBJECT", self.styles["line_label_center"]),
                    Paragraph("TYPE", self.styles["line_label_center"]),
                    Paragraph("DATE", self.styles["line_label_center"]),
                ]
            ]
            for box_item in box.boxitem_set.all().order_by("position"):
                aliquot = self.get_aliquot(box_item.identifier)
                panel_object = self.get_panel_object(aliquot)
                barcode = code39.Standard39(
                    aliquot.aliquot_identifier, barHeight=5 * mm, stop=1
                )
                table_data.append(
                    [
                        barcode,
                        Paragraph(str(box_item.position), self.styles["row_data"]),
                        Paragraph(aliquot.human_readable_identifier, self.styles["row_data"]),
                        Paragraph(aliquot.subject_identifier, self.styles["row_data"]),
                        Paragraph(
                            f"{aliquot.aliquot_type} ({aliquot.numeric_code}) {panel_object.abbreviation}",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            aliquot.aliquot_datetime.strftime("%Y-%m-%d"),
                            self.styles["row_data"],
                        ),
                    ]
                )
            t1 = Table(table_data)
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
                        ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)
        return story

    def get_aliquot(self, box_item_identifier=None):
        """Returns the aliquot instance for this box item."""
        try:
            aliquot = self.aliquot_model.objects.get(aliquot_identifier=box_item_identifier)
        except ObjectDoesNotExist as e:
            raise ManifestPdfReportError(
                f"{e} Got Box item '{box_item_identifier}'",
                code="invalid_aliquot_identifier",
            )
        return aliquot

    def get_panel_object(self, aliquot=None):
        """Returns the panel object associated with this
        aliquot.
        """
        requisition = None
        for requisition_model in site_labs.requisition_models:
            model_cls = django_apps.get_model(requisition_model)
            try:
                requisition = model_cls.objects.get(
                    requisition_identifier=aliquot.requisition_identifier
                )
            except ObjectDoesNotExist:
                pass
        if not requisition:
            raise ManifestPdfReportError(
                f"Invalid requisition identifier. Got {aliquot.requisition_identifier}",
                code="invalid_requisition_identifier",
            )
        return requisition.panel_object



from django import forms

from edc_constants.constants import YES

from ..diagnoses import (
    ClinicalReviewBaselineRequired,
    Diagnoses,
    InitialReviewRequired,
    MultipleInitialReviewsExist,
)


class DiagnosisFormValidatorMixin:
    def get_diagnoses(self) -> Diagnoses:
        try:
            diagnoses = Diagnoses(
                subject_identifier=self.subject_identifier,
                report_datetime=self.report_datetime,
            )
        except ClinicalReviewBaselineRequired as e:
            raise forms.ValidationError(e)
        try:
            diagnoses.get_initial_reviews()
        except InitialReviewRequired as e:
            raise forms.ValidationError(e)
        except MultipleInitialReviewsExist as e:
            raise forms.ValidationError(e)
        return diagnoses

    def applicable_if_not_diagnosed(
        self, diagnoses=None, prefix=None, field_applicable=None, label=None
    ) -> bool:
        diagnoses = diagnoses or self.get_diagnoses()
        return self.applicable_if_true(
            diagnoses.get_dx(prefix) != YES,
            field_applicable=field_applicable,
            applicable_msg=(
                f"Patient was not previously diagnosed with {label}. Expected YES or NO."
            ),
            not_applicable_msg=f"Patient was previously diagnosed with {label}.",
        )

    def applicable_if_diagnosed(
        self, diagnoses=None, prefix=None, field_applicable=None, label=None
    ) -> bool:
        diagnoses = diagnoses or self.get_diagnoses()
        diagnosed = diagnoses.get_dx(prefix) == YES
        return self.applicable_if_true(
            diagnosed,
            field_applicable=field_applicable,
            applicable_msg=(
                f"Patient was previously diagnosed with {label}. Expected YES or NO."
            ),
            not_applicable_msg=f"Patient was not previously diagnosed with {label}.",
        )



from django.apps import apps as django_apps
from edc_model_wrapper import ModelWrapper

from edc_lab.models import BoxItem

edc_lab_app_config = django_apps.get_app_config("edc_lab")


class BaseBoxItemModelWrapper(ModelWrapper):
    model_cls = BoxItem
    action_name = None
    next_url_name = None
    next_url_attrs = ["box_identifier", "action_name"]

    @property
    def human_readable_identifier(self):
        return self.object.human_readable_identifier

    @property
    def box_identifier(self):
        return self.object.box.box_identifier
